﻿using Janus.Core.Importers;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Janus.Analyzers.CircularDependency
{
    public class CircularDependencyAnalyzer
    {
        /// <summary>
        /// Dictionary of unique circles in the graph
        /// Stores the list of items in the circle
        /// Key:    Sorted ids of the dependencies in the circle as string
        /// Value:  List of the dependencies in the unique circle
        /// </summary>
        Dictionary<string, List<DependencyItem>> uniqueCircles = new Dictionary<string, List<DependencyItem>>();

        ////DFS fields

        /// <summary>
        /// To be analyzed stack (May be queue in the future ...)
        /// </summary>
        Stack<DependencyItem> dfsToBeAnalysed = new Stack<DependencyItem>();

        /// <summary>
        /// Current path in the dfs alg
        /// </summary>
        Stack<DependencyItem> dfsPath = new Stack<DependencyItem>();

        /// <summary>
        /// Current path in the dfs mirrored, for faster contains check
        /// </summary>
        HashSet<DependencyItem> dfsPathSetFast = new HashSet<DependencyItem>();

        //Key:   Visited dependency refference
        HashSet<DependencyItem> dfsVisitedDependencies = new HashSet<DependencyItem>();


        public CircularDependencyAnalyzer()
        {

        }

        /// <summary>
        /// Processes the second part of deserialization and redirects the analysis to the data specific methods
        /// </summary>
        /// <param name="dataWrapper">Data wrapper to be redirected</param>
        public void ProcessDataWrapper(DataSerializationWrapper dataWrapper)
        {
            if (dataWrapper.DataType == null)
            {
                //NULL!!!!
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is null, cannot deserialize!");
                return;
            }
            else if (dataWrapper.DataType == new List<DependencyItem>().GetType().ToString())
            {
                ImportResult<IList<DependencyItem>> importedData = JsonImporter.Instance.ImportData<IList<DependencyItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeDependencies(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<ProjectItem>().GetType().ToString())
            {
                ImportResult<IList<ProjectItem>> importedData = JsonImporter.Instance.ImportData<IList<ProjectItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeProjects(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<SolutionItem>().GetType().ToString())
            {
                ImportResult<IList<SolutionItem>> importedData = JsonImporter.Instance.ImportData<IList<SolutionItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeSolutions(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else
            {
                //Unknown ...
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is unknown, cannot deserialize!");
                return;
            }

            //Should be imported fully now ... start the analysis
            StartDependencyAnalysis();
        }


        /// <summary>
        /// Performs the analysis queue for each solution of the given list of the solutions
        /// Internally calls <see cref="AnalyzeSolution"/>
        /// </summary>
        /// <param name="solutions">Solutions to analyze</param>
        private void AnalyzeSolutions(IList<SolutionItem> solutions)
        {
            foreach (var solution in solutions)
            {
                AnalyzeSolution(solution);
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project in the given solution
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="solution">Solution to analyze</param>
        private void AnalyzeSolution(SolutionItem solution)
        {
            foreach (var projectEdge in solution.Projects)
            {
                if (projectEdge.Link != null)
                {
                    AnalyzeProject(projectEdge.Link);
                }
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project of the given list of the projects
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="projects">Projects to analyze</param>
        private void AnalyzeProjects(IList<ProjectItem> projects)
        {
            foreach (var project in projects)
            {
                AnalyzeProject(project);
            }
        }

        /// <summary>
        /// Enques the direct dependencies of the project to the work queue
        /// </summary>
        /// <param name="project">Projects to analyze</param>
        private void AnalyzeProject(ProjectItem project)
        {
            foreach (var dependencyEdge in project.Dependencies)
            {
                if (dependencyEdge.Link != null)
                {
                    dfsToBeAnalysed.Push(dependencyEdge.Link); //Add to the work list
                }
            }
        }

        /// <summary>
        /// Enques each dependency to the work queue
        /// </summary>
        /// <param name="dependencies">Dependencies to analyze</param>
        private void AnalyzeDependencies(IList<DependencyItem> dependencies)
        {
            foreach (var dependency in dependencies)
            {
                dfsToBeAnalysed.Push(dependency); //Add to the work list
                //AnalyzeDependency(dependency); 
            }
        }

        /// <summary>
        /// Performs the dependency analysis with the filled work queue of dpeendencies 
        /// </summary>
        private void StartDependencyAnalysis()
        {
            foreach (var rootNode in dfsToBeAnalysed)
            {
                DFS(rootNode);
            }

            //Finalize the obtained reults
            FinalizeAnalysis();
        }


        /// <summary>
        /// Annotates the given data with the cycle error mark
        /// Adds to the text the id+version of the other memebers in the cycle
        /// </summary>
        private void FinalizeAnalysis()
        {
            foreach (var uniqueCircle in uniqueCircles.Values)
            {
                string pathText = "";
                uniqueCircle.ForEach(x => {
                    pathText += $"({x.Id} {x.Version})->";
                });

                string cycleFoundText = $"Cycle found {pathText}";

                foreach (var node in uniqueCircle)
                {
                    node.AnnotationInfo.Add(new AnnotationInfoItem()
                    {
                        AnalyzerName = this.GetType().Name,
                        AnalyzerVersion = Assembly.GetEntryAssembly().GetName().Version,
                        AnalyzerCondition = $"Cycle",
                        InformationType = AnnotationInfoItem.InformationLevel.Err,
                        Text = cycleFoundText
                    });
                }
            }
        }

        //DFS

        /// <summary>
        /// DFS alg implementation with path preservation
        /// </summary>
        /// <param name="startItem">Item to be started with</param>
        private void DFS(DependencyItem startItem)
        {
            DFSMarkVisited(startItem);
            foreach (var item in startItem.Dependencies)
            {
                DependencyItem subdependency = item.Link;
                if (!dfsVisitedDependencies.Contains(subdependency))
                {                   
                    DFS(subdependency);     //New item not visited
                }
                else
                {
                    //OK Visited, is it in path? Yes = cycle!!!
                    if (dfsPathSetFast.Contains(subdependency))
                    {
                        var cycleFound = GetFullGraphCyclePath(subdependency);
                        InsertCycleListToDictionary(cycleFound);
                    }
                }
            }
            DFSRemoveFromPath(startItem);
        }

        /// <summary>
        /// Marks the item as visited and adds it to the path
        /// </summary>
        /// <param name="item">Item to be processed as visited</param>
        private void DFSMarkVisited(DependencyItem item)
        {
            dfsVisitedDependencies.Add(item);
            dfsPath.Push(item);
            dfsPathSetFast.Add(item);
        }

        /// <summary>
        /// Remove item from the path
        /// </summary>
        /// <param name="item">Item to be removed</param>
        private void DFSRemoveFromPath(DependencyItem item)
        {
            dfsPath.Pop();
            dfsPathSetFast.Remove(item);
        }

        /// <summary>
        /// Gets the full graph cycle path with the help of DFS memory
        /// </summary>
        /// <param name="cycleItem">Item which will create cycle</param>
        /// <returns>List of items which create cycle</returns>
        private List<DependencyItem> GetFullGraphCyclePath(DependencyItem cycleItem)
        {
            List<DependencyItem> cycleNodes = new List<DependencyItem>();
            var path = dfsPath.ToList();
            path.Reverse();
            bool occurenceFound = false;
            foreach (var node in path)
            {
                //Cycle is from subdepedency first occurence in the list to the end
                if (!occurenceFound && node == cycleItem)
                {
                    occurenceFound = true;
                }
                if (occurenceFound)
                {
                    cycleNodes.Add(node);
                }
            }
            return cycleNodes;
        }

        /// <summary>
        /// Inserts cycle to the list of found cycles, if unique
        /// </summary>
        /// <param name="cycleItems"></param>
        private void InsertCycleListToDictionary(List<DependencyItem> cycleItems)
        {
            //Create key
            string key = "";
            //Sort to get key 

            List<DependencyItem> clonedList = new List<DependencyItem>(cycleItems);
            clonedList.Sort();
            clonedList.ForEach(x => {
                key += $"{x.Id},{x.Version};";
            });

            if (!uniqueCircles.ContainsKey(key))
            {
                //New cycle -> insert
                uniqueCircles.Add(key, cycleItems);
            }
        }
    }
}
