﻿using CommandLine;
using Janus.Core.Importers;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Janus.Core.Utils;
using System;
using System.Collections.Generic;
using System.IO;

namespace Janus.Analyzers.KeyValue
{
    class Program
    {
        /// <summary>
        /// Command line arguments parsed as options
        /// </summary>
        private static CLIOptions _options;

        /// <summary>
        /// Main entry pojnt of the program
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            var result = CommandLine.Parser.Default.ParseArguments<CLIOptions>(args)
                                                            .WithParsed(RunOptions)
                                                            .WithNotParsed(HandleParseError);
        }

        /// <summary>
        /// Main logic, if argument parsing is without an error
        /// </summary>
        /// <param name="opts">CLI Arguments as options</param>
        /// <returns></returns>
        static void RunOptions(CLIOptions opts)
        {
            ImportResult<DataSerializationWrapper> importResult;
            _options = opts; //Pass args to the rest of the program

            IList<Tuple<string, string>> searchPatterns = new List<Tuple<string, string>>();
            foreach (var searchPattern in _options.SearchPatterns)
            {
                string[] parts = searchPattern.Split('=');

                if (parts.Length != 2)
                {
                    Console.Error.WriteLine($"ERROR: Search pattern in wrong format! Cannot parse: {searchPattern}");
                    return;
                }

                Tuple<string, string> tuppleTemp = new Tuple<string, string>(parts[0], parts[1]);
                searchPatterns.Add(tuppleTemp);
            }

            if (opts.InputFile != null)
            {
                //Try to read from file

                if (!File.Exists(opts.InputFile))
                {
                    Console.Error.WriteLine($"ERROR: File does not exist! (Path: {opts.InputFile})");
                    return;
                }

                //File exists, try to read

                importResult = JsonImporter.Instance.ImportWrapperFromFile(opts.InputFile);
            }
            else
            {
                //File input not specified, fallback to std in
                importResult = JsonImporter.Instance.ImportWrapperFromStream(Console.In);
            }
            //File loaded
            if (!importResult.OperationSuccessful)
            {
                //Import problem .... end execution
                foreach (var err in importResult.Errors)
                {
                    Console.Error.WriteLine(err);
                }
                return;
            }

            //OK Ready to process
            KeyValueAnalyzer kvAnalyzer = new KeyValueAnalyzer(searchPatterns);

            try
            {
                //Sanity check -> this will prevent from failing
                kvAnalyzer.ProcessDataWrapper(importResult.Result);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine($"ERROR: During analysis ...");
                Console.Error.WriteLine($"Detailed error: {e.Message}");
            }

            try
            {
                //Data processed -> Send to output
                DataWorker.Print(importResult.Result.Data, _options);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine($"ERROR: During data print ...");
                Console.Error.WriteLine($"Detailed error: {e.Message}");

            }
        }

        /// <summary>
        /// This method is called in case of argument parsing failure
        /// </summary>
        /// <param name="errs">Parsing errors as text</param>
        static void HandleParseError(IEnumerable<Error> errs)
        {
            Console.Error.WriteLine("ERROR During arguments matching ...");
            foreach (var item in errs)
            {
                Console.Error.WriteLine(item);
            }
        }
    }
}
