﻿using Janus.Core.Interfaces;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Janus.Core.Importers
{
    /// <summary>
    /// This class is used for data import of the json format.
    /// Works with the provided universal data format, stored in json.
    /// </summary>
    public class JsonImporter : IImporter
    {
        /// <summary>
        /// Private instance of the JsonImporter used as singleton
        /// </summary>
        private static JsonImporter _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private JsonImporter()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static JsonImporter Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new JsonImporter();
                }
                return _instance;
            }
        }

        #region Interface implementation

        /// <summary>
        /// Imports the wrapper from the given file
        /// Internaly calls the <see cref="ImportWrapperFromString"/>
        /// 
        /// Wrapper is not loaded completely, Data object has to be loaded with the coresponding datatype stored in the atribute of the Wrapper
        /// </summary>
        /// <param name="filePath">Path to the file, to be loaded</param>
        /// <param name="skipValidation">Skips the version validation</param>
        /// <returns>Import result, with the imported Wrapper object</returns>
        public ImportResult<DataSerializationWrapper> ImportWrapperFromFile(string filePath, bool skipValidation = false)
        {
            ImportResult<DataSerializationWrapper> result = new ImportResult<DataSerializationWrapper>();
            string jsonText = null;

            try
            {
                jsonText = File.ReadAllText(filePath);
            }
            catch (Exception e)
            {
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
                return result;
            }

            return ImportWrapperFromString(jsonText, skipValidation);
        }

        /// <summary>
        /// Imports the wrapper from the given data stream, has to be ended
        /// Internaly calls the <see cref="ImportWrapperFromString"/>
        /// 
        /// Wrapper is not loaded completely, Data object has to be loaded with the coresponding datatype stored in the atribute of the Wrapper
        /// </summary>
        /// <param name="stream">Datastream, to used as an input for characters</param>
        /// <param name="skipValidation">Skips the version validation</param>
        /// <returns>Import result, with the imported Wrapper object</returns>
        public ImportResult<DataSerializationWrapper> ImportWrapperFromStream(TextReader stream, bool skipValidation = false)
        {
            ImportResult<DataSerializationWrapper> result = new ImportResult<DataSerializationWrapper>();
            string streamData = "";

            try
            {
                streamData = stream.ReadToEnd();
            }
            catch (Exception e)
            {
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
                return result;
            }

            return ImportWrapperFromString(streamData, skipValidation);
        }

        /// <summary>
        /// Imports the wrapper from the given string
        /// Internaly calls the <see cref="ValidateWrapperVersion"/> to validate version of the loaded wrapper
        /// 
        /// Wrapper is not loaded completely, Data object has to be loaded with the coresponding datatype stored in the atribute of the Wrapper
        /// </summary>
        /// <param name="jsonData">Json data as text</param>
        /// <param name="skipValidation">Skips the version validation</param>
        /// <returns>Import result, with the imported Wrapper object</returns>
        public ImportResult<DataSerializationWrapper> ImportWrapperFromString(string jsonData, bool skipValidation = false)
        {
            ImportResult<DataSerializationWrapper> result = new ImportResult<DataSerializationWrapper>();

            try
            {
                result.Result = JsonConvert.DeserializeObject<DataSerializationWrapper>(jsonData,
                                                                                           new JsonSerializerSettings { PreserveReferencesHandling = PreserveReferencesHandling.Objects });

                if (!skipValidation)
                {
                    BaseResult validationResult = ValidateWrapperVersion(result.Result.Version);
                    if (!validationResult.OperationSuccessful)
                    {
                        //Validation failed
                        result.OperationSuccessful = false;
                        result.Errors = validationResult.Errors; //Forward errors
                        return result;
                    }
                }

                result.OperationSuccessful = true;
            }
            catch (Exception e)
            {
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
                return result;
            }

            return result;
        }

        /// <summary>
        /// Imports the data from the Wrapper using the type provided
        /// 
        /// </summary>
        /// <typeparam name="T">Type of the data in the Data atribute</typeparam>
        /// <param name="dataWrapper">Data wrapper to be used</param>
        /// <returns>Import result</returns>
        public ImportResult<T> ImportData<T>(DataSerializationWrapper dataWrapper)
        {
            ImportResult<T> result = new ImportResult<T>();

            try
            {
                result.Result = JsonConvert.DeserializeObject<T>(dataWrapper.Data.ToString(),
                                                           new JsonSerializerSettings
                                                           {
                                                               PreserveReferencesHandling = PreserveReferencesHandling.Objects,
                                                               MissingMemberHandling = MissingMemberHandling.Error
                                                           });
                result.OperationSuccessful = true;
            }
            catch (Exception e)
            {
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
                return result;
            }

            return result;
        }



        #endregion

        #region Private methods

        /// <summary>
        /// Performs the version validation, checks if the Majors are the same. Also checks if the Minor in core is greater or equeal the minor in the imported file.
        /// </summary>
        /// <param name="importDataversion">Version of the imported file</param>
        /// <returns>Result of the validation, optionaly with errors</returns>
        private BaseResult ValidateWrapperVersion(Version importDataversion)
        {
            BaseResult result = new BaseResult();
            Version currentCoreVersion = Janus.Core.Consts.UniversalDataFormatVersion;

            ////Semantic versioning 
            
            if (currentCoreVersion.Major > importDataversion.Major)
            {
                result.OperationSuccessful = false;
                result.Errors.Add("ERROR: Unable to import! Currently used core has higher Major version of the DataSerializationWrapper than imported file!");
                return result;
            }
            else if (currentCoreVersion.Major < importDataversion.Major)
            {
                result.OperationSuccessful = false;
                result.Errors.Add("ERROR: Unable to import! Imported file has higher Major version of the DataSerializationWrapper than currently used core!");
                return result;
            }

            //Majors OK, lets check minor

            if (currentCoreVersion.Minor < importDataversion.Minor)
            {
                result.OperationSuccessful = false;
                result.Errors.Add("ERROR: Unable to import! Imported file has higher Minor version of the DataSerializationWrapper than currently used core!");
                return result;
            }

            result.OperationSuccessful = true;

            return result;
        }

        #endregion

    }
}
