﻿using Janus.Core.Enums;
using Janus.Core.Exporters;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Text;

namespace Janus.Core.Utils
{
    /// <summary>
    /// This class is used for data printing operations
    /// This will redirect the output to the desired location based on the options
    /// </summary>
    public static class DataWorker
    {

        /// <summary>
        /// This method processes the given data to the desired output based on the options.
        /// </summary>
        /// <typeparam name="T">Data type</typeparam>
        /// <param name="dataToPrint">Data object to be printed</param>
        /// <param name="options">Options, containing the output specification</param>
        /// <returns>Operation result</returns>
        public static bool Print<T>(T dataToPrint, CLIOptionsCoreBase options)
        {
            switch (options.OutputType)
            {
                case EOutputType.JSON:
                    ExportResult jsonExportResult = JsonExporter.Instance.ExportData(dataToPrint, options.OutputFile);
                    return ProcessExportResult(jsonExportResult);
                case EOutputType.MD:
                    Console.Error.WriteLine("MD Export not implemented!");
                    break;
                case EOutputType.DOT:
                    ExportResult dotExportResult = DotExporter.Instance.ExportData(dataToPrint, options.OutputFile);
                    return ProcessExportResult(dotExportResult);
                case EOutputType.Text:
                    ExportResult textExportResult = TextExporter.Instance.ExportData(dataToPrint, options.OutputFile);
                    return ProcessExportResult(textExportResult);
                default:
                    Console.Error.WriteLine($"Unknown option to print, \"{options.OutputType}\" not implemented!");
                    break;
            }
            return false;
        }

        /// <summary>
        /// This method processes the ExportResult and performs additional operations if required
        /// Additional opration = print to std out if it wasnt printed to the file, using the exporter itself
        /// </summary>
        /// <param name="result">Exported data to be additionaly processed</param>
        /// <returns>Result of the export</returns>
        private static bool ProcessExportResult(ExportResult result)
        {
            if (result.OperationSuccessful)
            {
                //Succes

                if (!result.ExportedToFile)
                {
                    //Not to file ... export to std out
                    Console.WriteLine(result.ExportData);
                }

                return true;
            }
            else
            {
                //Failed ... print why

                Console.Error.WriteLine("Export failed!");
                foreach (var textError in result.Errors)
                {
                    Console.Error.WriteLine(textError);
                }

                return result.OperationSuccessful;
            }
        }

    }
}
