﻿using Janus.Core.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Janus.Core.Utils
{
    /// <summary>
    /// Utilities for the graph operations
    /// 
    /// Consist of useful extension methods
    /// </summary>
    public static class Utils
    {
        /// <summary>
        /// Performs action for each direct subdependency in dependencyItem
        /// </summary>
        /// <param name="item"></param>
        /// <param name="lambdaAction"></param>
        public static void ForEachDirectSubDependency(this DependencyItem item, Action<DependencyItem, DependencyEdge> lambdaAction)
        {
            foreach (var dependencyEdge in item.Dependencies)
            {
                if (dependencyEdge.Link != null)
                {
                    lambdaAction(dependencyEdge.Link, dependencyEdge);
                }
            }
        }

        /// <summary>
        /// Performs action for each direct subdependency in dependencyItem
        /// Using BFS algorithm
        /// </summary>
        /// <param name="item">Item to be started BFS on</param>
        /// <param name="lambdaAction">Provides 3 values in this sequence: Parrent item, edge, child item</param>
        public static void ForEachSubDependencyEdgeBfs(this DependencyItem item, Action<DependencyItem, DependencyEdge, DependencyItem> lambdaAction)
        {
            Queue<DependencyItem> bfsToBeAnalysed = new Queue<DependencyItem>();

            //Key:   Visited dependency refference
            HashSet<DependencyItem> bfsVisitedDependencies = new HashSet<DependencyItem>();

            //Start bfs
            bfsToBeAnalysed.Enqueue(item);

            while (bfsToBeAnalysed.Any())
            {
                DependencyItem current = bfsToBeAnalysed.Dequeue();

                if (bfsVisitedDependencies.Contains(current))
                {
                    //Already visited!!!
                    continue;
                }

                bfsVisitedDependencies.Add(current); //Mark as visited

                //Process dependency NODE
                //AnalyzeDependency(current);

                foreach (var subDependencyEdge in current.Dependencies)
                {
                    DependencyItem subDependency = subDependencyEdge.Link;
                    if (subDependency == null)
                    {
                        //SKIP null ...
                        continue;
                    }
                    if (bfsVisitedDependencies.Contains(subDependency))
                    {
                        //Already visited ... but the edge has to shown in the graph
                        lambdaAction(current, subDependencyEdge, subDependency);
                    }
                    else
                    {
                        //New dependency, has to be processed .. or already in the queue?
                        if (bfsToBeAnalysed.Contains(subDependency))
                        {
                            //Is included in to do list ... just reference .. but has to be shown in the graph
                            lambdaAction(current, subDependencyEdge, subDependency);
                        }
                        else
                        {
                            //Completely new ... will be included in queue
                            lambdaAction(current, subDependencyEdge, subDependency);
                            bfsToBeAnalysed.Enqueue(subDependency);        //Add to the to process queue

                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets the list of the unique dependencies and the unique edges for the given dependency acting as the root
        /// </summary>
        /// <param name="item">Dependency to be selected as the root</param>
        /// <returns>Tuple, first item is the list of unique dependencies, second item is the list of unique edges</returns>
        public static Tuple<List<DependencyItem>, List<DependencyEdge>> GetUniqueNodesAndEdges(this DependencyItem item)
        {
            HashSet<DependencyItem> deps = new HashSet<DependencyItem>();
            HashSet<DependencyEdge> edges = new HashSet<DependencyEdge>();

            //Process the BFS
            item.ForEachSubDependencyEdgeBfs((source, edge, child) =>
            {
                deps.Add(source);
                deps.Add(child);
                edges.Add(edge);
            });

            return Tuple.Create(deps.ToList(), edges.ToList());
        }
    }
}
