﻿using Janus.Core.Models;
using Microsoft.Build.Construction;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Linq;

namespace Janus.Parsers.DotnetCoreNuget.Parser
{
    /// <summary>
    /// This class helps with the parsing of the Dotnet Core projects and solutions
    /// </summary>
    class DotnetCoreParser
    {
        /// <summary>
        /// Private constructor, this should not be instanced
        /// </summary>
        private DotnetCoreParser()
        {

        }

        /// <summary>
        /// This method parses the solution and returns the SolutionItem object, filled with all information available about child projects and its direct dependencies.
        /// </summary>
        /// <param name="solutionPath">Path to the solution file</param>
        /// <returns>SolutionItem object, filled with all information about its child projects</returns>
        public static SolutionItem ParseSolution(string solutionPath)
        {
            var solutionFile = SolutionFile.Parse(solutionPath);

            var mySolution = new SolutionItem(solutionPath) { Language = "C#"};

            //Console.WriteLine();
            //Console.WriteLine("Found projects in solution: ");
            foreach (var item in solutionFile.ProjectsInOrder)
            {
                if(item.ProjectType != SolutionProjectType.SolutionFolder)
                {
                    ProjectItem projItem = new ProjectItem(item.AbsolutePath);
                    mySolution.Projects.Add(new ProjectEdge() { Link = projItem });
                    //Console.WriteLine("|---" + item.AbsolutePath);
                }
            }

            //Console.WriteLine();
            //Console.WriteLine("Parsing project dependencies...");
            foreach(var item in mySolution.Projects)
            {
                //Console.WriteLine("|------" + item.Path);
                if(item.Link != null)
                {
                    ParseProject(item.Link);
                }
                /*foreach (var dependency in item.Dependencies)
                {
                    Console.WriteLine($"     |--- {dependency.Id} (Version: {dependency.Version})");

                }*/
            }

            return mySolution;
        }

        /// <summary>
        /// This method creates the project item and calls the parsing method for project
        /// </summary>
        /// <param name="csprojPath">Path to the project</param>
        /// <returns>ProjectItem object, filled with all information about its dependencies</returns>
        public static ProjectItem ParseProject(string csprojPath)
        {
            ProjectItem project = new ProjectItem(csprojPath);
            return ParseProject(project);
        }

        /// <summary>
        /// This method parses the project and returns the ProjectItem object, filled with all information available about  its direct dependencies.
        /// </summary>
        /// <param name="project">Project item to be examined</param>
        /// <returns>ProjectItem object, filled with all information about its dependencies</returns>
        public static ProjectItem ParseProject(ProjectItem project)
        {
            XElement projectNode = XElement.Load(project.Path);
            //XNamespace ns = "http://schemas.microsoft.com/developer/msbuild/2003"; //Might be useful for the document
            var targetVersionNode = projectNode.Descendants("PropertyGroup").Descendants("TargetFramework"); //Obtains the target version
            var referenceNodes = projectNode.Descendants("ItemGroup").Descendants("PackageReference");       //Obtains the packages

            project.Language = "C#";

            if(targetVersionNode?.First() != null)
            {
                //Version node has some targets, it can be passed to the LanguageVersion property
                project.LanguageVersion = targetVersionNode.First().Value;
            }

            //Loop trough the dependencies
            foreach (var property in referenceNodes)
            {
                var includeAttr = property.Attribute("Include");
                var versionAttr = property.Attribute("Version");
                if (includeAttr != null && versionAttr != null)
                {
                    DependencyItem depItem = NugetAPI.NugetWrapper.Instance.FillMetadataAsync(includeAttr.Value, versionAttr.Value, Program.options.RepositoryURL).Result; //TODO: Fix this ugly repo url call

                    var versionLimit = new VersionLimit() { RequiredVersion = versionAttr.Value, LimitSymbol = VersionLimit.OperatorType.Equal };
                    var depEdge = new DependencyEdge() { Link = depItem };
                    depEdge.VersionLimits.Add(versionLimit);

                    project.Dependencies.Add(depEdge);
                }

                //Console.WriteLine($"Name: Include - Value: {property.Attribute("Include").Value}");
                //Console.WriteLine($"Name: Version - Value: {property.Attribute("Version")}");
            }

            return project;
        }
    }
}
